/*
 * mpcut - X11 mp3 frame editor
 *
 * Copyright (c) 2000, Warren Toomey, wkt@cs.adfa.edu.au
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Version 1.1b, $Revision: 1.21 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

extern GtkObject *adj1;			/* Adjustment for scroll bar */
static GtkAdjustment *adj=NULL;

/* Complicated dialog boxes can be slow to create, so we create them once and
 *  keep pointers to them in static variables so we can reuse them.
 */
static GtkWidget *open_filesel = NULL;
static GtkWidget *save_filesel = NULL;
static GtkWidget *hruler = NULL;

static GdkPixmap *pixmap = NULL;	/* Backing pixmap for drawing area */
static gfloat skrollval = 0.0;		/* Scrollbar at the bottom */

static GtkWidget *drawing_area = NULL;	/* Drawing area */
static int drawheight, drawwidth=0;	/* Current drawing area size */
static int startcol = 0;		/* Left-most frame number */
static int endcol = 0;			/* Right-most frame number */
static int framemark = -1;		/* Start of marked area */

FILE *infile = NULL;			/* Input file */

					/* Mpeg details from input file */
static int numframes;			/* # frames in input file */
static int hipower;			/* biggest energy level so far */

					/* Selection details */
int is_selection = 0;			/* If 1, there is a selection */
int select_start, select_end;		/* Frame# of start, end of selection */

static GdkColor Red, Blue;

/* Static functions. */
static void open_file(GtkWidget * main_window);
static void save_file(GtkWidget * main_window);
void real_open_file(GtkWidget * main_window, gchar * filename);
static void real_save_file(GtkWidget * main_window, gchar * filename);
static void draw_powerlines(void);

/* Functions in mpglib.c */
int get_maxframe(FILE * zin);
int get_maxframepower(FILE * zin);
int get_framelevel(FILE * zin, int framenum);
long get_frameoffset(FILE * zin, int framenum);
void ExitMP3(void);

/* Functions in deletions.c */
int is_frame_deleted(int frame);
void insert_deletion(int start, int end);
void clear_deletions(void);
void remove_first_deletion(void);

/* A key used to store pointers to the main window. */
static const gchar *MainWindowKey = "MainWindowKey";


void on_file1_activate(GtkMenuItem * menuitem,
		        gpointer user_data)
{

}


void on_open1_activate(GtkMenuItem * menuitem,
		        gpointer user_data)
{
  GtkWidget *main_window;

  /*
   * We use the Glade utility function lookup_widget() to get a pointer to the
   * main window. The first argument is any widget in the window/dialog and the
   * second argument is the name of the widget you want to get.
   */
  main_window = lookup_widget(GTK_WIDGET(menuitem), "main_window");
  open_file(main_window);
}


void on_save_as1_activate(GtkMenuItem * menuitem,
			   gpointer user_data)
{
  GtkWidget *main_window;

  main_window = lookup_widget(GTK_WIDGET(menuitem), "main_window");
  save_file(main_window);
}


void on_quit1_activate(GtkMenuItem * menuitem,
		        gpointer user_data)
{
  gtk_exit(0);
}


void on_select1_activate(GtkMenuItem * menuitem,
			  gpointer user_data)
{
}


void on_select_all1_activate(GtkMenuItem * menuitem,
			      gpointer user_data)
{
  is_selection = 1;
  select_start = 0, select_end = numframes - 1;
  draw_powerlines();
}


void on_unselect_all1_activate(GtkMenuItem * menuitem,
			        gpointer user_data)
{
  is_selection = 0;
  draw_powerlines();
}


void on_help1_activate(GtkMenuItem * menuitem,
		        gpointer user_data)
{

}


void on_about_mpcut1_activate(GtkMenuItem * menuitem,
			       gpointer user_data)
{
  GtkWidget *about;

  about = create_about_window();
  gtk_widget_show(about);

}


void on_mpcut_documentation1_activate(GtkMenuItem * menuitem,
				       gpointer user_data)
{

}


/* This is called when the main window gets a "delete_event" signal, which
 * typically happens when the user clicks on the close icon on the window's
 * title bar. If we didn't handle this, the window would be destroyed but
 *  our application would not exit.
 */
gboolean
on_mainwin_delete_event(GtkWidget * widget,
			GdkEvent * event,
			gpointer user_data)
{
  gtk_exit(0);
  /* Shouldn't reach here, but we add it to stop compiler warnings. */
  return FALSE;
}


void on_about_clicked(GtkButton * button,
		       gpointer user_data)
{
  gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


/* This shows the file selection dialog to open a file. */
static void open_file(GtkWidget * main_window)
{
  /*
   * We use the same file selection widget each time, so first of all we create
   * it if it hasn't already been created.
   */
  if (open_filesel == NULL) open_filesel = create_openfile();

  /*
   * We save a pointer to the main window inside the file selection's data
   * list, so we can get it easily in the callbacks.
   */
  gtk_object_set_data(GTK_OBJECT(open_filesel), MainWindowKey, main_window);

  gtk_widget_show(open_filesel);
  gdk_window_raise(open_filesel->window);
}

/* This shows the file selection dialog to save a file. */
static void save_file(GtkWidget * main_window)
{
  /*
   * We use the same file selection widget each time, so first of all we create
   * it if it hasn't already been created.
   */
  if (save_filesel == NULL) save_filesel = create_savefile();

  /*
   * We save a pointer to the main window inside the file selection's data
   * list, so we can get it easily in the callbacks.
   */
  gtk_object_set_data(GTK_OBJECT(save_filesel), MainWindowKey, main_window);

  gtk_widget_show(save_filesel);
  gdk_window_raise(save_filesel->window);
}

void on_open_cancel__clicked(GtkButton * button,
			      gpointer user_data)
{
  gtk_widget_hide(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


void on_save_cancel_clicked(GtkButton * button,
			     gpointer user_data)
{
  gtk_widget_hide(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}

/* Create a new backing pixmap of the appropriate size */
gint on_drawing_configure(GtkWidget * widget,
			   GdkEventConfigure * event)
{
  GdkColormap *colourmap;
  if (hruler==NULL) hruler = lookup_widget(GTK_WIDGET(widget), "hruler1");

  				/* Add colours to the drawing area */
  colourmap = gdk_window_get_colormap(widget->window);
  Red.red= 65535; Red.green=0; Red.blue=0;
  gdk_color_alloc(colourmap,&Red);
  Blue.red= 0; Blue.green=0; Blue.blue=65535;
  gdk_color_alloc(colourmap,&Blue);

  if (pixmap) gdk_pixmap_unref(pixmap);

  pixmap = gdk_pixmap_new(widget->window, widget->allocation.width,
			  widget->allocation.height, -1);
  drawheight = widget->allocation.height;
  drawwidth = widget->allocation.width;
  drawing_area = widget;

  adj=NULL;			/* Force new scrollbar values */

  if (infile == NULL) gdk_draw_rectangle(pixmap, widget->style->white_gc,
		       		TRUE, 0, 0, drawwidth, drawheight);
  return TRUE;
}

void draw_powerlines(void)
{
  GdkGC *gc;
  float scalefactor;
  int i, x, level;

  if (infile == NULL) return;		/* Nothing to draw, return */

  if (adj1 && (adj==NULL)) {
    /* Set up the dimensions of the scrollbar */
    adj=GTK_ADJUSTMENT(adj1);
    /* adj->value=0.0; */
    adj->page_size= (gfloat)drawwidth / (gfloat)numframes;
    adj->step_increment= 0.25 * adj->page_size;
    adj->page_increment= 0.5 * adj->page_size;
    adj->upper= 1.0 + adj->page_increment;
    gtk_adjustment_changed(adj);
  }

  /* Work out the starting position of the first column */
  if (numframes <= drawwidth) startcol = 0;
  else startcol = (int) (skrollval * (float) numframes);

  /* Now work out the end column */
  endcol = startcol + drawwidth - 1;
  if (endcol >= numframes) endcol = numframes - 1;

  /* Get the level for the right-most frame; this causes
   * the input file to be read if it hasn't done so already
   */
  level = get_framelevel(infile, endcol);
  if (level == -1) {
    fprintf(stderr, "Error reading file\n"); return;
  }
  /* Now find the highest power so far */
  hipower = get_maxframepower(infile);
  if (hipower == -1) {
    fprintf(stderr, "Error reading file\n"); return;
  }
  /* Work out the scale factor */
  scalefactor = (float) drawheight / (float) hipower;

  /* Clear the drawing area */
  gdk_draw_rectangle(pixmap, drawing_area->style->white_gc,
		     TRUE, 0, 0, drawwidth, drawheight);

  /* Set the ruler */
  if (hruler)
     gtk_ruler_set_range(GTK_RULER(hruler),
	(gfloat)startcol, (gfloat)endcol, (gfloat)startcol, (gfloat)endcol);

  /* Ok, let's start reading frame powers, scale and draw them */
  for (x = 0, i = startcol; i <= endcol; x++, i++) {
    level = get_framelevel(infile, i);
    level = (int) ((float) level * scalefactor);

    if (i==framemark) {
      gc = gdk_gc_new(drawing_area->window); gdk_gc_set_foreground(gc,&Blue);
      gdk_draw_line(pixmap, gc, x, drawheight, x, 0);

    } else if (is_frame_deleted(i)) {
      gdk_draw_line(pixmap, drawing_area->style->mid_gc[0], 
		    x, drawheight, x, drawheight - level);

    } else if (is_selection && (i >= select_start) && (i <= select_end)) {
      gc = gdk_gc_new(drawing_area->window); gdk_gc_set_foreground(gc,&Red);
      gdk_draw_line(pixmap, gc, x, drawheight, x, drawheight - level);

    } else
      gdk_draw_line(pixmap, drawing_area->style->black_gc,
		    x, drawheight, x, drawheight - level);
  }
  gdk_draw_pixmap(drawing_area->window,
		  drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
		  pixmap, 0, 0, 0, 0, drawwidth, drawheight);
}


/* Redraw the screen from the backing pixmap */
gint on_drawing_expose(GtkWidget * widget,
		        GdkEventExpose * event)
{
  if (infile==NULL) gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)], pixmap,
		  event->area.x, event->area.y,
		  event->area.x, event->area.y,
		  event->area.width, event->area.height);
  else draw_powerlines();
  return FALSE;
}

void on_open_ok(GtkButton * button,
		 gpointer user_data)
{
  GtkWidget *filesel, *main_window;
  gchar *filename;

  filesel = gtk_widget_get_toplevel(GTK_WIDGET(button));
  main_window = gtk_object_get_data(GTK_OBJECT(filesel), MainWindowKey);
  gtk_widget_hide(filesel);
  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  real_open_file(main_window, filename);
}

/* This loads the given file. */
void real_open_file(GtkWidget * main_window, gchar * filename)
{
  FILE *fp;

  fp = fopen(filename, "r");
  if (fp == NULL) {
    fprintf(stderr, "Couldn't open file %s\n", filename); return;
  }
  if (infile) {
    ExitMP3(); fclose(infile);
    skrollval = 0.0;
    startcol = endcol = is_selection = 0;
  }
  infile = fp;

  /* Get the number of frames */
  numframes = get_maxframe(infile);
  if (numframes == -1) {
    fprintf(stderr, "Error reading file %s\n", filename); return;
  }

  framemark=-1;			/* No blue frame mark */
  clear_deletions();		/* Clear any existing deletions */
				/* Ensure scrollbar starts at zero */
  if (adj1) { adj=GTK_ADJUSTMENT(adj1);adj->value=0.0; }
  adj=NULL;			/* Force new scrollbar values */
  if (drawwidth>0) draw_powerlines();
}

#ifndef min
#define min(x,y) ( (x<y) ? x : y )
#endif
#define BUFSIZE 8192

/* Given an open FILE pointer, and the start and end frame
 * numbers, write the frames to the file. Close the FILE
 * pointer before we return. We skip those frames which
 * are marked as being deleted.
 */
static void write_selection(FILE *out, int start_frame, int end_frame)
{
  int frame;
  long i, old_offset, start_offset, end_offset, amount, bufamount;
  char buf[BUFSIZE];

		/* Save current offset before we move around in infile */
  old_offset = ftell(infile);

  frame=start_frame;
  while (1) {
				/* Find 1st non-deleted frame */
     while ((frame<=end_frame) && is_frame_deleted(frame)) frame++;
     if (frame>end_frame) break;
     start_offset = get_frameoffset(infile, frame++);

				/* Find next non-deleted frame */
     while ((frame<=end_frame) && !is_frame_deleted(frame)) frame++;
     end_offset = get_frameoffset(infile, frame);

     if ((start_offset == -1) || (end_offset == -1)) {	/* Bad offset #s */
	fclose(out); return;
     }

				/* Work out how much to write */
     amount = end_offset - start_offset;
     fseek(infile, start_offset, SEEK_SET);

				/* And read/write buffers until done */
     while (amount>0) {
       bufamount = min(amount, BUFSIZE);
       i = fread(buf, 1, bufamount, infile);
       if (i < bufamount) break;
       fwrite(buf, 1, i, out);
       amount -= i;
     }
  }
  /* Go back to the position in the file before we started writing */
  fclose(out); fseek(infile, old_offset, SEEK_SET);
}

/* This saves the selection. */
static void real_save_file(GtkWidget * main_window, gchar * filename)
{
  FILE *fp;

  if (is_selection == 0) return;		/* Nothing to save */

  fp = fopen(filename, "w");
  if (fp == NULL) {
    fprintf(stderr, "Couldn't save file %s\n", filename); return;
  }
  write_selection(fp, select_start, select_end);
}


void on_play_selection1_activate(GtkMenuItem * menuitem,
				  gpointer user_data)
{
  FILE *fp;
  int i, pid, status;
  int pfd[2];

  if (is_selection == 0) return;		/* Nothing to save */

  i = pipe(pfd);				/* Make a pipe */
  if (i == -1) return;

  pid = fork();
  switch (pid) {
    case -1: break;
    case 0:					/* Child */
      close(pfd[1]);				/* Close fds not required */
      close(0); fclose(infile);
      dup2(pfd[0], 0);
      close(pfd[0]);				/* Make pipe the stdin */
      execlp("mpg123", "mpg123", "-q", "-", NULL);
      fprintf(stderr, "Hmm, execlp of mpg123 failed\n");
      exit(1);
    default:					/* Parent */
      close(pfd[0]);
      signal(SIGPIPE, SIG_IGN);			/* Don't die on pipe errs */
      fp= fdopen(pfd[1], "w");
      if (fp!=NULL) write_selection(fp, select_start, select_end);
      close(pfd[1]);
					/* Reap the exit status of the child */
      waitpid(pid, &status, 0);
  }
}


void on_save_ok(GtkButton * button,
		 gpointer user_data)
{
  GtkWidget *filesel, *main_window;
  gchar *filename;

  filesel = gtk_widget_get_toplevel(GTK_WIDGET(button));
  main_window = gtk_object_get_data(GTK_OBJECT(filesel), MainWindowKey);
  gtk_widget_hide(filesel);
  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  real_save_file(main_window, filename);
}

gboolean
on_skroll_changed(GtkWidget * widget,
		  GdkEventButton * event,
		  gpointer user_data)
{
  GtkAdjustment *adjustment;
  gfloat oldval;

  oldval = skrollval;
  adjustment = (GtkAdjustment *) user_data;
  skrollval = adjustment->value;
  if (oldval != skrollval) draw_powerlines();
  return FALSE;
}

#define DRAW_BORD 8		/* Size of drawing border */
static drag_start = -1;

/* This function is called either on a single-click or at the
 * end of a mouse drag event. If it's a drag, drag_start != -1.
 * We do `mark' events on middle buttons, normal selections on
 * other buttons.
 */
gboolean
on_draw_click(GtkWidget * widget,
	      GdkEventButton * event,
	      gpointer user_data)
{
  int start, end, swap;
  int first_in = 0, second_in = 0;

  /* Save the end of a drag, but make sure it fits within the drawing area. */
  end = (int) event->x - DRAW_BORD;
  if (end < 0) end = 0;
  if (end >= drawwidth) end = drawwidth - 1;

  /* Get start of drag. */
  start = drag_start; drag_start = -1;

  /* Middle button - do mark handling */
  if (event->button==2) {
     if (start==-1) {
      	framemark=end+startcol;		/* Click: mark the frame */
        draw_powerlines();
        return FALSE;
     } else {
	if (framemark==-1) return FALSE;	/* No mark set yet! */
	start= framemark;		/* Drag: selection is from the */
	end += startcol;		/* mark to the drag's end */
  	if (end > endcol) end = endcol; 
  	goto set_selection;		/* Yuk, a goto */
     }
  }

  /* Must be some other button */
  /* If not a drag, do an unselection */
  if (start == -1) {
    if (is_selection == 1) { is_selection = 0; draw_powerlines(); }
    return FALSE;
  }

  /* Event was a drag! Convert both start and end to frame numbers */
  /* by adding the startcol */
  start += startcol;
  end += startcol;
  if (end > endcol) end = endcol;

  /* Find out if either endpoints are in the current selection */
  if (is_selection) {
    if ((start >= select_start) && (start <= select_end)) first_in = 1;
    if ((end >= select_start) && (end <= select_end)) second_in = 1;
  }
  /* Both in or both out, a new selection */
  if (first_in == second_in) goto set_selection;	/* Yuk, a goto */

  /* First point in, extend the selection */
  if (first_in) {
    if (end < select_start) { start = end; end = select_end; }
    else start = select_start;
  } else {
    /* Last point in, shrink the selection */
    if (start < select_start) { start = end; end = select_end; }
    else start = select_start;
  }

set_selection:
  if (end < start) { swap = end; end = start; start = swap; }
  select_start = start; select_end = end; is_selection = 1;
  draw_powerlines();
  return FALSE;
}

gboolean
on_draw_drag(GtkWidget * widget,
	     GdkEventMotion * event,
	     gpointer user_data)
{
  /* Save start of a drag, & make sure it fits within the drawing area. */
  drag_start = (int) event->x - DRAW_BORD;
  if (drag_start < 0) drag_start = 0;
  if (drag_start >= drawwidth) drag_start = drawwidth - 1;
  return FALSE;
}

void
on_cut_selection(GtkMenuItem *menuitem,
		gpointer user_data)
{
 if (is_selection==0) return;
 insert_deletion(select_start, select_end);
 draw_powerlines();
}

void
on_remove_cuts(GtkMenuItem *menuitem,
               gpointer user_data)
{
 clear_deletions();
 draw_powerlines();
}

void
on_undo_cut(GtkMenuItem *menuitem,
               gpointer user_data)
{
 remove_first_deletion();
 draw_powerlines();
}

